#ifdef GL_ES
precision mediump float;
#endif

uniform sampler2D tex0;
uniform vec2 size; // Size of the texture in pixels
//uniform int frame; // Number of elapsed frames
uniform bool evenframe; // Is this frame even or odd?
uniform int orientation;

uniform sampler2D noisetex;
uniform mat4 noisemat;
uniform vec2 noisesize;

varying vec2 tcoord;

// Material IDs
// Materials are arranged in order of density, so that density can be compared using < or >

const int FIRE = 0;
const int STEAM = 1;
const int AIR = 2;
const int WATER = 3;
const int SAND = 4;
const int WALL = 5;
const int PLANT = 6;
const int SMOLDER = 7;
const int NUM_MATERIALS = 8;

// A random value that will be different for each 2x2 block of cells
float randval;

void swap(inout int a, inout int b)
{
	int t = a;
	a = b;
	b = t;
}

// Swap two cells if at least one of them matches type, and the other is of a less dense type
void swapif(int type, inout int a, inout int b)
{
	if (a == type || b == type) if (b < type || a < type) swap(a, b);
}

void change(int typeA, int typeB, inout int a, inout int b, inout int c, inout int d)
{
	if (a == typeA) a = typeB;
	if (b == typeA) b = typeB;
	if (c == typeA) c = typeB;
	if (d == typeA) d = typeB;
}

void randchange(int typeA, int typeB, float odds, inout int a, inout int b, inout int c, inout int d)
{
	if (a == typeA && randval < odds) a = typeB;
	if (b == typeA && randval >= odds && randval < odds * 2.0) b = typeB;
	if (c == typeA && randval >= odds * 2.0 && randval < odds * 3.0) c = typeB;
	if (d == typeA && randval >= odds * 3.0 && randval < odds * 4.0) d = typeB;
}

// Get the material type of the cell at the offset position
int get(vec2 offset)
{
	vec2 pos = tcoord + offset / size;
	//if (pos.y > 1.0) return WALL;
	//else if (pos.y < 0.0) return AIR;
	//pos.x = mod(pos.x, 1.0);
	pos = mod(pos, vec2(1, 1));
	return int(texture2D(tex0, pos).r * float(NUM_MATERIALS - 1) + 0.5);
}

void gravitylogic(inout int a, inout int b, inout int c, inout int d)
{
	// Move denser materials (but not walls) down, displacing less dense materials
	if (a < WALL && c < a) swap(a, c);
	if (b < WALL && d < b) swap(b, d);
}

void sandlogic(inout int a, inout int b, inout int c, inout int d)
{
	// Spreads out more in the air than otherwise
	bool spreadInAir = (randval < 0.1 && c != SAND && d != SAND);
	if (spreadInAir || randval < 0.01) swapif(SAND, a, b);
	if (randval < 0.9) swapif(SAND, c, d); // Forms diagonal slopes
}

void waterlogic(inout int a, inout int b, inout int c, inout int d)
{
	// Spreads out more on the surface than otherwise
	bool spreadOnSurface = ((c == WATER || d == WATER) && randval < 0.99);
	if (spreadOnSurface || randval < 0.05) swapif(WATER, a, b);
	if (randval < 0.98 || (a < WATER || b < WATER)) swapif(WATER, c, d);
}

void plantlogic(inout int a, inout int b, inout int c, inout int d)
{
	bool nearPlant = (a == PLANT || b == PLANT || c == PLANT || d == PLANT);
	bool nearAir = (a == AIR || b == AIR || c == AIR || d == AIR);
	if (nearPlant && nearAir) change(WATER, PLANT, a, b, c, d);
}

void firelogic(inout int a, inout int b, inout int c, inout int d)
{
	randchange(STEAM, WATER, 0.001, a, b, c, d);
	randchange(FIRE, STEAM, 0.05, a, b, c, d);
	randchange(SMOLDER, FIRE, 0.1, a, b, c, d);

	bool nearFire = a == FIRE || b == FIRE || c == FIRE || d == FIRE;
	//if (randval > 0.9) nearFire = nearFire || a == SMOLDER || b == SMOLDER || c == SMOLDER || d == SMOLDER;
	if (nearFire)
	{
		bool nearWater = a == WATER || b == WATER || c == WATER || d == WATER;

		change(PLANT, SMOLDER, a, b, c, d);
		change(WATER, STEAM, a, b, c, d);

		if (nearWater) change(FIRE, STEAM, a, b, c, d);
	}
}

void airlogic(inout int a, inout int b, inout int c, inout int d)
{
	if (randval < 0.5) 
	{
		swapif(AIR, a, b);
		swapif(AIR, c, d);
		swapif(STEAM, a, b);
		swapif(STEAM, c, d);
	}

	if (randval < 0.1 || randval > 0.9)
	{
		swapif(AIR, a, c);
		swapif(AIR, b, d);
		swapif(STEAM, a, c);
		swapif(STEAM, b, d);
	}
}

void main()
{
	vec2 pixelcoord = floor(tcoord * size);
	vec2 offset = mod(pixelcoord + float(evenframe), 2.0);
	vec2 gridcoord = (pixelcoord - offset) / 2.0;

	vec2 noisecoord = mod((noisemat * vec4(gridcoord, 0, 1)).xy / noisesize, vec2(1, 1));

	vec3 noisesample = texture2D(noisetex, noisecoord).rgb;

	randval = noisesample.r * 0.99 + noisesample.g * 0.0099 + noisesample.b * 0.000099;

	//randval = rand(vec2(gridcoord.x, (mod(gridcoord.y, 512.0) + 1.0) * (mod(float(frame), 512.0) + 1.0)));

	ivec4 C = ivec4(
		get((vec2(0, 0) - offset)), 
		get((vec2(1, 0) - offset)),
		get((vec2(0, 1) - offset)),
		get((vec2(1, 1) - offset))
	);

	if (orientation == 1) C = C.zxwy;
	else if (orientation == 2) C = C.wzyx;
	else if (orientation == 3) C = C.ywxz;
	
	waterlogic(C.x, C.y, C.z, C.w);
	sandlogic(C.x, C.y, C.z, C.w);
	plantlogic(C.x, C.y, C.z, C.w);
	gravitylogic(C.x, C.y, C.z, C.w);
	firelogic(C.x, C.y, C.z, C.w);
	airlogic(C.x, C.y, C.z, C.w);
	
	if (orientation == 1) C = C.ywxz;
	else if (orientation == 2) C = C.wzyx;
	else if (orientation == 3) C = C.zxwy;

	int outvalue = C.x * int(offset == vec2(0, 0))
				 + C.y * int(offset == vec2(1, 0))
				 + C.z * int(offset == vec2(0, 1))
				 + C.w * int(offset == vec2(1, 1));
	
	gl_FragColor = vec4(vec3(float(outvalue) / float(NUM_MATERIALS - 1)), 1);

	//gl_FragColor = vec4(texture2D(noisetex, tcoord).rgb, 1);
}